@@ -1,5 +1,5 @@
 /* libinfinity - a GObject-based infinote implementation
- * Copyright (C) 2007-2011 Armin Burgmeier <armin@arbur.net>
+ * Copyright (C) 2007-2014 Armin Burgmeier <armin@arbur.net>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -19,11 +19,10 @@
 
 #include <infinoted/infinoted-run.h>
 #include <infinoted/infinoted-dh-params.h>
-#include <infinoted/infinoted-creds.h>
 #include <infinoted/infinoted-util.h>
-#include <infinoted/infinoted-note-plugin.h>
 
 #include <libinfinity/server/infd-filesystem-storage.h>
+#include <libinfinity/server/infd-filesystem-account-storage.h>
 #include <libinfinity/server/infd-tcp-server.h>
 #include <libinfinity/common/inf-standalone-io.h>
 #include <libinfinity/common/inf-discovery-avahi.h>
@@ -40,17 +39,35 @@ infinoted_run_load_directory(InfinotedRun* run,
                              InfinotedStartup* startup,
                              GError** error)
 {
-  /* TODO: Allow different storage plugins */
   InfdFilesystemStorage* storage;
+  InfdFilesystemAccountStorage* account_storage;
   InfCommunicationManager* communication_manager;
 
 #ifdef G_OS_WIN32
   gchar* module_path;
 #endif
   gchar* plugin_path;
+  gboolean result;
 
   storage = infd_filesystem_storage_new(startup->options->root_directory);
 
+  /* Set an account storage, too. This can be replaced by something more fancy
+   * by a plugin. */
+  account_storage = infd_filesystem_account_storage_new();
+
+  result = infd_filesystem_account_storage_set_filesystem(
+    account_storage,
+    storage,
+    error
+  );
+
+  if(result == FALSE)
+  {
+    g_object_unref(account_storage);
+    g_object_unref(storage);
+    return FALSE;
+  }
+
   communication_manager = inf_communication_manager_new();
 
   run->io = inf_standalone_io_new();
@@ -61,39 +78,56 @@ infinoted_run_load_directory(InfinotedRun* run,
     communication_manager
   );
 
+  g_object_set(
+    G_OBJECT(run->directory),
+    "account-storage",
+    account_storage,
+    NULL
+  );
+
   infd_directory_enable_chat(run->directory, TRUE);
 
   g_object_unref(storage);
+  g_object_unref(account_storage);
   g_object_unref(communication_manager);
 
+  /* Load server plugins via plugin manager */
 #ifdef G_OS_WIN32
   module_path = g_win32_get_package_installation_directory_of_module(NULL);
-  plugin_path = g_build_filename(module_path, "lib", PLUGIN_BASEPATH, NULL);
+  plugin_path = g_build_filename(module_path, "lib", PLUGIN_PATH, NULL);
   g_free(module_path);
 #else
-  plugin_path = g_build_filename(PLUGIN_LIBPATH, PLUGIN_BASEPATH, NULL);
+  plugin_path = g_build_filename(PLUGIN_LIBPATH, PLUGIN_PATH, NULL);
 #endif
 
-  if(!infinoted_note_plugin_load_directory(plugin_path, run->directory))
-  {
-    g_free(plugin_path);
+  run->plugin_manager = infinoted_plugin_manager_new(
+    run->directory,
+    startup->log,
+    startup->credentials
+  );
 
+  result = infinoted_plugin_manager_load(
+    run->plugin_manager,
+    plugin_path,
+    (const gchar* const*)startup->options->plugins,
+    startup->options->config_key_file,
+    error
+  );
+
+  g_free(plugin_path);
+  infinoted_options_drop_config_file(startup->options);
+
+  if(result == FALSE)
+  {
+    g_object_unref(run->plugin_manager);
     g_object_unref(run->directory);
     g_object_unref(run->io);
+    run->plugin_manager = NULL;
     run->directory = NULL;
     run->io = NULL;
-
-    g_set_error(
-      error,
-      g_quark_from_static_string("INFINOTED_STARTUP_ERROR"),
-      0,
-      "Failed to load note plugins"
-    );
-
     return FALSE;
   }
 
-  g_free(plugin_path);
   return TRUE;
 }
 
@@ -225,34 +259,6 @@ infinoted_run_new(InfinotedStartup* startup,
 
   inf_ip_address_free(address);
 
-  run->record = infinoted_record_new(run->directory);
-
-  if(startup->options->autosave_interval > 0)
-  {
-    run->autosave = infinoted_autosave_new(
-      run->directory,
-      startup->options->autosave_interval
-    );
-  }
-  else
-  {
-    run->autosave = NULL;
-  }
-
-  if(startup->options->sync_interval > 0 &&
-     startup->options->sync_directory != NULL)
-  {
-    run->dsync = infinoted_directory_sync_new(
-      run->directory,
-      startup->options->sync_directory,
-      startup->options->sync_interval
-    );
-  }
-  else
-  {
-    run->dsync = NULL;
-  }
-
   return run;
 }
 
@@ -270,11 +276,6 @@ infinoted_run_free(InfinotedRun* run)
   if(inf_standalone_io_loop_running(run->io))
     inf_standalone_io_loop_quit(run->io);
 
-  if(run->autosave != NULL)
-    infinoted_autosave_free(run->autosave);
-  if(run->dsync != NULL)
-    infinoted_directory_sync_free(run->dsync);
-
   if(run->xmpp6 != NULL)
   {
     g_object_get(G_OBJECT(run->xmpp6), "status", &status, NULL);
@@ -297,8 +298,11 @@ infinoted_run_free(InfinotedRun* run)
   g_object_unref(run->avahi);
 #endif
 
-  if(run->record != NULL)
-    infinoted_record_free(run->record);
+  if(run->plugin_manager != NULL)
+  {
+    g_object_unref(run->plugin_manager);
+    run->plugin_manager = NULL;
+  }
 
   g_object_unref(run->io);
   g_object_unref(run->directory);
@@ -340,13 +344,16 @@ infinoted_run_start(InfinotedRun* run)
   if(run->startup->credentials)
   {
     result = infinoted_dh_params_ensure(
-      run->startup->credentials, &run->dh_params, &error);
+      run->startup->log, run->startup->credentials, &run->dh_params, &error);
 
     if(result == FALSE)
     {
-      infinoted_util_log_error(
-             _("Failed to generate Diffie-Hellman parameters: %s"),
-             error->message);
+      infinoted_log_error(
+        run->startup->log,
+        _("Failed to generate Diffie-Hellman parameters: %s"),
+        error->message
+      );
+
       g_error_free(error);
       return;
     }
@@ -360,7 +367,12 @@ infinoted_run_start(InfinotedRun* run)
     if(infd_tcp_server_open(tcp, &error6) == TRUE)
     {
       g_object_get(G_OBJECT(tcp), "local-port", &port, NULL);
-      infinoted_util_log_info(_("IPv6 Server running on port %u"), port);
+
+      infinoted_log_info(
+        run->startup->log,
+        _("IPv6 Server running on port %u"),
+        port
+      );
     }
     else
     {
@@ -378,7 +390,12 @@ infinoted_run_start(InfinotedRun* run)
     if(infd_tcp_server_open(tcp, &error4) == TRUE)
     {
       g_object_get(G_OBJECT(tcp), "local-port", &port, NULL);
-      infinoted_util_log_info(_("IPv4 Server running on port %u"), port);
+
+      infinoted_log_info(
+        run->startup->log,
+        _("IPv4 Server running on port %u"),
+        port
+      );
     }
     else
     {
@@ -394,9 +411,12 @@ infinoted_run_start(InfinotedRun* run)
   {
     g_assert(error4 != NULL || error6 != NULL);
     error = error4 != NULL ? error4 : error6;
-    infinoted_util_log_error(
-            _("Failed to start server: %s"),
-            error->message);
+
+    infinoted_log_error(
+      run->startup->log,
+      _("Failed to start server: %s"),
+      error->message
+    );
   }
 
   if(error4 != NULL) g_error_free(error4);
@@ -407,7 +427,14 @@ infinoted_run_start(InfinotedRun* run)
   fflush(stderr);
 
   if(run->xmpp4 != NULL || run->xmpp6 != NULL)
+  {
     inf_standalone_io_loop(run->io);
+
+    infinoted_log_info(
+      run->startup->log,
+      _("Infinoted shutting down...")
+    );
+  }
 }
 
 /**
