@@ -1,5 +1,5 @@
 /* libinfinity - a GObject-based infinote implementation
- * Copyright (C) 2007-2011 Armin Burgmeier <armin@arbur.net>
+ * Copyright (C) 2007-2014 Armin Burgmeier <armin@arbur.net>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -19,7 +19,7 @@
 
 #include <infinoted/infinoted-startup.h>
 #include <infinoted/infinoted-util.h>
-#include <infinoted/infinoted-creds.h>
+#include <infinoted/infinoted-log.h>
 
 #include <infinoted/infinoted-pam.h>
 
@@ -39,13 +39,14 @@ infinoted_startup_free_certificate_array(gnutls_x509_crt_t* certificates,
                                          guint n_certificates)
 {
   guint i;
-  for(i = 0; i < n_certificates; ++ i)
+  for(i = 0; i < n_certificates; ++i)
     gnutls_x509_crt_deinit(certificates[i]);
   g_free(certificates);
 }
 
 static gnutls_x509_privkey_t
-infinoted_startup_load_key(gboolean create_key,
+infinoted_startup_load_key(InfinotedLog* log,
+                           gboolean create_key,
                            const gchar* key_file,
                            GError** error)
 {
@@ -58,13 +59,13 @@ infinoted_startup_load_key(gboolean create_key,
 
     /* TODO: Open the key file beforehand */
 
-    infinoted_util_log_info(_("Generating 2048 bit RSA private key..."));
-    key = infinoted_creds_create_key(error);
+    infinoted_log_info(log, _("Generating 4096 bit RSA private key..."));
+    key = inf_cert_util_create_private_key(GNUTLS_PK_RSA, 4096, error);
 
     if(key == NULL)
       return NULL;
 
-    if(infinoted_creds_write_key(key, key_file, error) == FALSE)
+    if(inf_cert_util_write_private_key(key, key_file, error) == FALSE)
     {
       gnutls_x509_privkey_deinit(key);
       return NULL;
@@ -72,35 +73,44 @@ infinoted_startup_load_key(gboolean create_key,
   }
   else
   {
-    key = infinoted_creds_read_key(key_file, error);
+    key = inf_cert_util_read_private_key(key_file, error);
   }
 
   return key;
 }
 
 static gnutls_x509_crt_t*
-infinoted_startup_load_certificate(gboolean create_self_signed_certificate,
+infinoted_startup_load_certificate(InfinotedLog* log,
+                                   gboolean create_self_signed_certificate,
                                    gnutls_x509_privkey_t key,
                                    const gchar* certificate_file,
                                    const gchar* certificate_chain_file,
                                    guint* n_certificates,
                                    GError** error)
 {
+  InfCertUtilDescription desc;
   gnutls_x509_crt_t* result;
   gnutls_x509_crt_t cert;
   GPtrArray* certs;
   GPtrArray* chain_certs;
+  gboolean res;
 
   if(create_self_signed_certificate == TRUE)
   {
     if(infinoted_util_create_dirname(certificate_file, error) == FALSE)
       return NULL;
 
-    infinoted_util_log_info(_("Generating self-signed certificate..."));
-    cert = infinoted_creds_create_self_signed_certificate(key, error);
+
+    infinoted_log_info(log, _("Generating self-signed certificate..."));
+    desc.validity = 365 * 24 * 3600;
+    desc.dn_common_name = g_get_host_name();
+    desc.san_dnsname = g_get_host_name();
+
+    cert = inf_cert_util_create_self_signed_certificate(key, &desc, error);
     if(cert == NULL) return NULL;
 
-    if(inf_cert_util_save_file(&cert, 1, certificate_file, error) == FALSE)
+    res = inf_cert_util_write_certificate(&cert, 1, certificate_file, error);
+    if(res == FALSE)
     {
       gnutls_x509_crt_deinit(cert);
       return NULL;
@@ -114,13 +124,13 @@ infinoted_startup_load_certificate(gboolean create_self_signed_certificate,
   }
   else
   {
-    certs = inf_cert_util_load_file(certificate_file, NULL, error);
+    certs = inf_cert_util_read_certificate(certificate_file, NULL, error);
     if(certs == NULL) return NULL;
 
     if(certificate_chain_file != NULL)
     {
       chain_certs =
-        inf_cert_util_load_file(certificate_chain_file, certs, error);
+        inf_cert_util_read_certificate(certificate_chain_file, certs, error);
 
       if(chain_certs == NULL)
       {
@@ -141,10 +151,16 @@ static gboolean
 infinoted_startup_load_credentials(InfinotedStartup* startup,
                                    GError** error)
 {
+  gnutls_x509_crt_t* certs;
+  guint n_certs;
+  gnutls_certificate_credentials_t creds;
+  int res;
+
   if(startup->options->security_policy !=
      INF_XMPP_CONNECTION_SECURITY_ONLY_UNSECURED)
   {
     startup->private_key = infinoted_startup_load_key(
+      startup->log,
       startup->options->create_key,
       startup->options->key_file,
       error
@@ -153,27 +169,37 @@ infinoted_startup_load_credentials(InfinotedStartup* startup,
     if(startup->private_key == NULL)
       return FALSE;
 
-    startup->certificates = infinoted_startup_load_certificate(
+    certs = infinoted_startup_load_certificate(
+      startup->log,
       startup->options->create_certificate,
       startup->private_key,
       startup->options->certificate_file,
       startup->options->certificate_chain_file,
-      &startup->n_certificates,
+      &n_certs,
       error
     );
 
-    if(startup->certificates == NULL)
+    if(certs == NULL)
       return FALSE;
 
-    startup->credentials = infinoted_creds_create_credentials(
-      startup->private_key,
-      startup->certificates,
-      startup->n_certificates,
-      error
+    /* Takes ownership of certificates: */
+    startup->certificates = inf_certificate_chain_new(certs, n_certs);
+
+    startup->credentials = inf_certificate_credentials_new();
+    creds = inf_certificate_credentials_get(startup->credentials);
+
+    res = gnutls_certificate_set_x509_key(
+      creds,
+      inf_certificate_chain_get_raw(startup->certificates),
+      inf_certificate_chain_get_n_certificates(startup->certificates),
+      startup->private_key
     );
 
-    if(startup->credentials == NULL)
+    if(res != 0)
+    {
+      inf_gnutls_set_error(error, res);
       return FALSE;
+    }
   }
 
   return TRUE;
@@ -206,12 +232,12 @@ infinoted_startup_load_options(InfinotedStartup* startup,
 
   config_files = g_malloc( (n_system_config_dirs + 2) * sizeof(gchar*));
   config_files[n_system_config_dirs + 1] = NULL;
-  config_files[n_system_config_dirs] =
+  config_files[0] =
     g_build_filename(user_config_dir, "infinoted.conf", NULL);
 
   for(i = 0; i < n_system_config_dirs; ++ i)
   {
-    config_files[n_system_config_dirs - i - 1] =
+    config_files[i + 1] =
       g_build_filename(system_config_dirs[i], "infinoted.conf", NULL);
   }
 
@@ -261,30 +287,39 @@ infinoted_startup_sasl_callback(InfSaslContextSession* session,
                                 gpointer user_data)
 {
   InfinotedStartup* startup;
+  const char* username;
   const char* password;
   InfXmppConnection* xmpp;
 
 #ifdef LIBINFINITY_HAVE_PAM
-  const char* username;
   const gchar* pam_service;
   GError* error;
 #endif
+  gchar* remote_id;
 
   xmpp = INF_XMPP_CONNECTION(session_data);
+  g_object_get(xmpp, "remote-id", &remote_id, NULL);
 
   switch(prop)
   {
   case GSASL_VALIDATE_SIMPLE:
     startup = (InfinotedStartup*)user_data;
+    username = inf_sasl_context_session_get_property(session, GSASL_AUTHID);
     password = inf_sasl_context_session_get_property(session, GSASL_PASSWORD);
 #ifdef LIBINFINITY_HAVE_PAM
-    username = inf_sasl_context_session_get_property(session, GSASL_AUTHID);
     pam_service = startup->options->pam_service;
     if(pam_service != NULL)
     {
       error = NULL;
       if(!infinoted_pam_authenticate(pam_service, username, password))
       {
+        infinoted_log_warning(
+          startup->log,
+          _("User %s failed to log in from %s: PAM authentication failed"),
+          username,
+          remote_id
+        );
+
         infinoted_startup_sasl_callback_set_error(
           xmpp,
           INF_AUTHENTICATION_DETAIL_ERROR_AUTHENTICATION_FAILED,
@@ -296,10 +331,15 @@ infinoted_startup_sasl_callback(InfSaslContextSession* session,
           GSASL_AUTHENTICATION_ERROR
         );
       }
-      else if(!infinoted_pam_user_is_allowed(startup->options,
-                                             username,
-                                             &error))
+      else if(!infinoted_pam_user_is_allowed(startup, username, &error))
       {
+        infinoted_log_warning(
+          startup->log,
+          _("User %s failed to log in from %s: PAM user not allowed"),
+          username,
+          remote_id
+        );
+
         infinoted_startup_sasl_callback_set_error(
           xmpp,
           INF_AUTHENTICATION_DETAIL_ERROR_USER_NOT_AUTHORIZED,
@@ -313,6 +353,13 @@ infinoted_startup_sasl_callback(InfSaslContextSession* session,
       }
       else
       {
+        infinoted_log_info(
+          startup->log,
+          _("User %s logged in from %s via PAM"),
+          username,
+          remote_id
+        );
+
         inf_sasl_context_session_continue(session, GSASL_OK);
       }
     }
@@ -322,10 +369,24 @@ infinoted_startup_sasl_callback(InfSaslContextSession* session,
       g_assert(startup->options->password != NULL);
       if(strcmp(startup->options->password, password) == 0)
       {
+        infinoted_log_info(
+          startup->log,
+          _("User %s logged in from %s via password"),
+          username,
+          remote_id
+        );
+
         inf_sasl_context_session_continue(session, GSASL_OK);
       }
       else
       {
+        infinoted_log_warning(
+          startup->log,
+          _("User %s failed to log in from %s: wrong password"),
+          username,
+          remote_id
+        );
+
         infinoted_startup_sasl_callback_set_error(
           xmpp,
           INF_AUTHENTICATION_DETAIL_ERROR_AUTHENTICATION_FAILED,
@@ -344,6 +405,8 @@ infinoted_startup_sasl_callback(InfSaslContextSession* session,
     inf_sasl_context_session_continue(session, GSASL_AUTHENTICATION_ERROR);
     break;
   }
+
+  g_free(remote_id);
 }
 
 static gboolean
@@ -357,6 +420,10 @@ infinoted_startup_load(InfinotedStartup* startup,
   if(infinoted_startup_load_options(startup, argc, argv, error) == FALSE)
     return FALSE;
 
+  startup->log = infinoted_log_new();
+  if(!infinoted_log_open(startup->log, startup->options->log_path, error))
+    return FALSE;
+
   if(infinoted_startup_load_credentials(startup, error) == FALSE)
     return FALSE;
 
@@ -403,9 +470,9 @@ infinoted_startup_new(int* argc,
 
   startup = g_slice_new(InfinotedStartup);
   startup->options = NULL;
+  startup->log = NULL;
   startup->private_key = NULL;
   startup->certificates = NULL;
-  startup->n_certificates = 0;
   startup->credentials = NULL;
   startup->sasl_context = NULL;
 
@@ -427,20 +494,20 @@ infinoted_startup_new(int* argc,
 void
 infinoted_startup_free(InfinotedStartup* startup)
 {
+  guint i;
+
   if(startup->credentials != NULL)
     inf_certificate_credentials_unref(startup->credentials);
 
   if(startup->certificates != NULL)
-  {
-    infinoted_startup_free_certificate_array(
-      startup->certificates,
-      startup->n_certificates
-    );
-  }
+    inf_certificate_chain_unref(startup->certificates);
 
   if(startup->private_key != NULL)
     gnutls_x509_privkey_deinit(startup->private_key);
 
+  if(startup->log != NULL)
+    g_object_unref(startup->log);
+
   if(startup->options != NULL)
     infinoted_options_free(startup->options);
 
