@@ -1,5 +1,5 @@
 /* libinfinity - a GObject-based infinote implementation
- * Copyright (C) 2007-2011 Armin Burgmeier <armin@arbur.net>
+ * Copyright (C) 2007-2014 Armin Burgmeier <armin@arbur.net>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -17,13 +17,25 @@
  * MA 02110-1301, USA.
  */
 
+/**
+ * SECTION:infinoted-util
+ * @title: InfinotedUtil
+ * @short_description: Miscellaneous helper functions.
+ * @include: infinoted/infinoted-util.h
+ * @stability: Unstable
+ *
+ * This section contains a few helper functions that are used in the
+ * infinoted implementation and are exposed to the plugin interface for
+ * convenience of plugin developers.
+ */
+
 #include <infinoted/infinoted-util.h>
 
+#include <libinfinity/common/inf-file-util.h>
 #include <libinfinity/inf-i18n.h>
 #include <libinfinity/inf-config.h>
 
 #ifdef LIBINFINITY_HAVE_LIBDAEMON
-#include <libdaemon/dlog.h>
 #include <libdaemon/dpid.h>
 
 #include <unistd.h> /* for access(2) */
@@ -39,43 +51,6 @@
 #include <stdarg.h>
 #include <stdio.h>
 
-#ifdef G_OS_WIN32
-/* Arbitrary; they are not used currently anyway */
-# define LOG_ERR 0
-# define LOG_WARNING 1
-# define LOG_INFO 2
-# include <windows.h>
-#else
-# include <syslog.h>
-#endif
-
-static void
-infinoted_util_logv(int prio, const char* fmt, va_list ap)
-{
-#ifdef LIBINFINITY_HAVE_LIBDAEMON
-  daemon_logv(prio, fmt, ap);
-#else
-#ifdef G_OS_WIN32
-  /* On Windows, convert to the character set of the console */
-  gchar* out;
-  gchar* codeset;
-  gchar* converted;
-
-  out = g_strdup_vprintf(fmt, ap);
-  codeset = g_strdup_printf("CP%u", (guint)GetConsoleOutputCP());
-  converted = g_convert(out, -1, codeset, "UTF-8", NULL, NULL, NULL);
-  g_free(out);
-  g_free(codeset);
-
-  fprintf(stderr, "%s\n", converted);
-  g_free(converted);
-#else
-  vfprintf(stderr, fmt, ap);
-  fputc('\n', stderr);
-#endif /* !G_OS_WIN32 */
-#endif /* !LIBINFINITY_HAVE_LIBDAEMON */
-}
-
 #ifdef LIBINFINITY_HAVE_LIBDAEMON
 static const gchar*
 infinoted_util_get_pidfile_path_user(void) {
@@ -112,80 +87,13 @@ infinoted_util_create_dirname(const gchar* path,
                               GError** error)
 {
   gchar* dirname;
-  int save_errno;
+  gboolean result;
 
   dirname = g_path_get_dirname(path);
-
-  if(g_mkdir_with_parents(dirname, 0700) != 0)
-  {
-    save_errno = errno;
-
-    g_set_error(
-      error,
-      g_quark_from_static_string("ERRNO_ERROR"),
-      save_errno,
-      _("Could not create directory \"%s\": %s"),
-      dirname,
-      strerror(save_errno)
-    );
-
-    g_free(dirname);
-    return FALSE;
-  }
-
+  result = inf_file_util_create_directory(dirname, 0700, error);
   g_free(dirname);
-  return TRUE;
-}
 
-/**
- * infinoted_util_log_error:
- * @fmt: A printf-style format string.
- * ...: Format arguments.
- *
- * Logs an error message. If the server is daemonized, log to syslog,
- * otherwise log to stderr.
- */
-void
-infinoted_util_log_error(const char* fmt, ...)
-{
-  va_list ap;
-  va_start(ap, fmt);
-  infinoted_util_logv(LOG_ERR, fmt, ap);
-  va_end(ap);
-}
-
-/**
- * infinoted_util_log_warning:
- * @fmt: A printf-style format string.
- * ...: Format arguments.
- *
- * Logs a warning message. If the server is daemonized, log to syslog,
- * otherwise log to stderr.
- */
-void
-infinoted_util_log_warning(const char* fmt, ...)
-{
-  va_list ap;
-  va_start(ap, fmt);
-  infinoted_util_logv(LOG_WARNING, fmt, ap);
-  va_end(ap);
-}
-
-/**
- * infinoted_util_log_info:
- * @fmt: A printf-style format string.
- * ...: Format arguments.
- *
- * Logs an info message. If the server is daemonized, log to syslog,
- * otherwise log to stderr.
- */
-void
-infinoted_util_log_info(const char* fmt, ...)
-{
-  va_list ap;
-  va_start(ap, fmt);
-  infinoted_util_logv(LOG_INFO, fmt, ap);
-  va_end(ap);
+  return result;
 }
 
 /**
@@ -225,7 +133,6 @@ infinoted_util_set_errno_error(GError** error,
   }
 }
 
-#ifdef LIBINFINITY_HAVE_LIBDAEMON
 
 /**
  * infinoted_util_daemon_set_global_pid_file_proc:
@@ -235,11 +142,13 @@ infinoted_util_set_errno_error(GError** error,
 void
 infinoted_util_daemon_set_global_pid_file_proc(void)
 {
+#ifdef LIBINFINITY_HAVE_LIBDAEMON
   daemon_pid_file_proc = infinoted_util_get_pidfile_path_system;
+#endif
 }
 
 /**
- * infinoted_util_daemon_set_global_pid_file_proc:
+ * infinoted_util_daemon_set_local_pid_file_proc:
  *
  * When attempting to read or write the PID file use the local file which is
  * in the owner's home directory.
@@ -247,7 +156,9 @@ infinoted_util_daemon_set_global_pid_file_proc(void)
 void
 infinoted_util_daemon_set_local_pid_file_proc(void)
 {
+#ifdef LIBINFINITY_HAVE_LIBDAEMON
   daemon_pid_file_proc = infinoted_util_get_pidfile_path_user;
+#endif
 }
 
 /**
@@ -262,13 +173,15 @@ infinoted_util_daemon_set_local_pid_file_proc(void)
 int
 infinoted_util_daemon_pid_file_kill(int sig)
 {
-#ifdef DAEMON_PID_FILE_KILL_WAIT_AVAILABLE
+#ifdef LIBINFINITY_HAVE_LIBDAEMON
+# ifdef DAEMON_PID_FILE_KILL_WAIT_AVAILABLE
   return daemon_pid_file_kill_wait(sig, 5);
-#else
+# else
   return daemon_pid_file_kill(sig);
+# endif
+#else
+  return 0;
 #endif
 }
 
-#endif
-
 /* vim:set et sw=2 ts=2: */
