@@ -1,5 +1,5 @@
 /* libinfinity - a GObject-based infinote implementation
- * Copyright (C) 2007-2011 Armin Burgmeier <armin@arbur.net>
+ * Copyright (C) 2007-2014 Armin Burgmeier <armin@arbur.net>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -17,10 +17,32 @@
  * MA 02110-1301, USA.
  */
 
-#include <libinfinity/client/infc-session-proxy.h>
-#include <libinfinity/client/infc-user-request.h>
-#include <libinfinity/client/infc-request-manager.h>
+/**
+ * SECTION:infc-session-proxy
+ * @title: InfcSessionProxy
+ * @short_description: Client-side session management
+ * @include: libinfinity/client/infc-session-proxy.h
+ * @see_also: #InfSession
+ * @stability: Unstable
+ *
+ * #InfcSessionProxy is a manager for #InfSession that is specific for the
+ * client side. The #InfcSessionProxy manages exactly one #InfSession. It is
+ * usually created by #InfcBrowser when a subscription to a node is made.
+ *
+ * #InfcSessionProxy allows to perform client-specific operations for the
+ * session it manages, such as joining a user. In addition it handles requests
+ * made by its counterpart, #InfdSessionProxy, for the server side, such as
+ * remote user joins and forced session closure.
+ *
+ * #InfcSessionProxy implements the #InfSessionProxy interface, which can be
+ * used to access the underlying #InfSession or to join a user.
+ */
 
+#include <libinfinity/client/infc-session-proxy.h>
+#include <libinfinity/client/infc-request-manager.h>
+#include <libinfinity/common/inf-session-proxy.h>
+#include <libinfinity/common/inf-session.h>
+#include <libinfinity/common/inf-request-result.h>
 #include <libinfinity/common/inf-xml-connection.h>
 #include <libinfinity/common/inf-xml-util.h>
 #include <libinfinity/common/inf-error.h>
@@ -266,6 +288,7 @@ infc_session_proxy_release_connection(InfcSessionProxy* proxy)
 
   g_assert(priv->connection != NULL);
   g_assert(priv->subscription_group != NULL);
+  g_assert(priv->request_manager != NULL);
 
   /* TODO: Emit failed signal with some "cancelled" error? */
   infc_request_manager_clear(priv->request_manager);
@@ -274,7 +297,8 @@ infc_session_proxy_release_connection(InfcSessionProxy* proxy)
   priv->request_manager = NULL;
 
   /* Set status of all users to unavailable */
-  /* TODO: Keep local users available */
+  /* TODO: Keep local users available if session is still open. Then make
+   * sure when the session is closed everybody is set to unavailable. */
   inf_user_table_foreach_user(
     inf_session_get_user_table(priv->session),
     infc_session_proxy_release_connection_foreach_user_func,
@@ -287,28 +311,40 @@ infc_session_proxy_release_connection(InfcSessionProxy* proxy)
     proxy
   );
 
-  inf_session_set_subscription_group(priv->session, NULL);
-
   g_object_unref(priv->subscription_group);
   priv->subscription_group = NULL;
 
   g_object_unref(G_OBJECT(priv->connection));
   priv->connection = NULL;
 
+  /* Keep the proxy alive while re-setting the session's subscription group,
+   * to be able to announce our status changes afterwards. */
+  g_object_ref(proxy);
+
+  inf_session_set_subscription_group(priv->session, NULL);
+
   g_object_notify(G_OBJECT(proxy), "connection");
   g_object_notify(G_OBJECT(proxy), "subscription-group");
+
+  g_object_unref(proxy);
 }
 
+/* TODO: This function should be moved to InfcRequest */
 static xmlNodePtr
 infc_session_proxy_request_to_xml(InfcRequest* request)
 {
   xmlNodePtr xml;
+  gchar* type;
+  guint seq;
   gchar seq_buffer[16];
 
-  xml = xmlNewNode(NULL, (const xmlChar*)infc_request_get_name(request));
-  sprintf(seq_buffer, "%u", infc_request_get_seq(request));
+  g_object_get(G_OBJECT(request), "type", &type, "seq", &seq, NULL);
+  sprintf(seq_buffer, "%u", seq);
 
+  xml = xmlNewNode(NULL, (const xmlChar*)type);
   xmlNewProp(xml, (const xmlChar*)"seq", (const xmlChar*)seq_buffer);
+
+  g_free(type);
   return xml;
 }
 
@@ -505,6 +541,7 @@ infc_session_proxy_handle_user_join(InfcSessionProxy* proxy,
   GArray* array;
   InfUser* user;
   GParameter* param;
+  gboolean result;
   guint i;
 
   priv = INFC_SESSION_PROXY_PRIVATE(proxy);
@@ -519,8 +556,7 @@ infc_session_proxy_handle_user_join(InfcSessionProxy* proxy,
 
   array = session_class->get_xml_user_props(priv->session, connection, xml);
 
-  /* Set local flag if the join was requested by us (seq is present in
-   * server response). */
+  /* Set local flag if the join was requested by us. */
   param = inf_session_get_user_property(array, "flags");
   g_assert(!G_IS_VALUE(&param->value)); /* must not have been set already */
 
@@ -538,14 +574,25 @@ infc_session_proxy_handle_user_join(InfcSessionProxy* proxy,
     g_value_set_object(&param->value, G_OBJECT(connection));
   }
 
-  /* This validates properties */
-  user = inf_session_add_user(
+  result = session_class->validate_user_props(
     priv->session,
     (const GParameter*)array->data,
     array->len,
+    NULL,
     error
   );
 
+  if(result == TRUE)
+  {
+    user = inf_session_add_user(
+      priv->session,
+      (const GParameter*)array->data,
+      array->len
+    );
+
+    g_assert(user != NULL);
+  }
+
   for(i = 0; i < array->len; ++ i)
     g_value_unset(&g_array_index(array, GParameter, i).value);
   g_array_free(array, TRUE);
@@ -554,8 +601,11 @@ infc_session_proxy_handle_user_join(InfcSessionProxy* proxy,
   {
     if(request != NULL)
     {
-      infc_user_request_finished(INFC_USER_REQUEST(request), user);
-      infc_request_manager_remove_request(priv->request_manager, request);
+      infc_request_manager_finish_request(
+        priv->request_manager,
+        request,
+        inf_request_result_make_join_user(INF_SESSION_PROXY(proxy), user)
+      );
     }
 
     return TRUE;
@@ -634,8 +684,7 @@ infc_session_proxy_handle_user_rejoin(InfcSessionProxy* proxy,
     goto error;
   }
 
-  /* Set local flag if the join was requested by us (seq is present in
-   * server response). */
+  /* Set local flag if the join was requested by us */
   param = inf_session_get_user_property(array, "flags");
   g_assert(!G_IS_VALUE(&param->value)); /* must not have been set already */
 
@@ -688,8 +737,11 @@ infc_session_proxy_handle_user_rejoin(InfcSessionProxy* proxy,
 
   if(request != NULL)
   {
-    infc_user_request_finished(INFC_USER_REQUEST(request), user);
-    infc_request_manager_remove_request(priv->request_manager, request);
+    infc_request_manager_finish_request(
+      priv->request_manager,
+      request,
+      inf_request_result_make_join_user(INF_SESSION_PROXY(proxy), user)
+    );
   }
 
   return TRUE;
@@ -832,8 +884,7 @@ infc_session_proxy_communication_object_enqueued(InfCommunicationObject* obj,
 static InfCommunicationScope
 infc_session_proxy_communication_object_received(InfCommunicationObject* obj,
                                                  InfXmlConnection* connection,
-                                                 xmlNodePtr node,
-                                                 GError** error)
+                                                 xmlNodePtr node)
 {
   InfcSessionProxy* proxy;
   InfcSessionProxyPrivate* priv;
@@ -859,8 +910,7 @@ infc_session_proxy_communication_object_received(InfCommunicationObject* obj,
     return inf_communication_object_received(
       INF_COMMUNICATION_OBJECT(priv->session),
       connection,
-      node,
-      error
+      node
     );
   }
   else
@@ -903,11 +953,11 @@ infc_session_proxy_communication_object_received(InfCommunicationObject* obj,
     }
     else
     {
+      /* forward to session */
       return inf_communication_object_received(
         INF_COMMUNICATION_OBJECT(priv->session),
         connection,
-        node,
-        error
+        node
       );
     }
   }
@@ -945,11 +995,66 @@ infc_session_proxy_communication_object_received(InfCommunicationObject* obj,
       g_error_free(seq_error);
     }
 
-    g_propagate_error(error, local_error);
+    /* TODO: Do we want to emit session's error signal? */
+    g_error_free(local_error);
   }
 
   /* Don't forward any of the handled messages */
-  return FALSE;
+  return INF_COMMUNICATION_SCOPE_PTP;
+}
+
+/*
+ * InfSessionProxy implementation
+ */
+
+InfRequest*
+infc_session_proxy_session_proxy_join_user(InfSessionProxy* proxy,
+                                           guint n_params,
+                                           const GParameter* params,
+                                           InfRequestFunc func,
+                                           gpointer user_data)
+{
+  InfcSessionProxyPrivate* priv;
+  InfSessionClass* session_class;
+  InfSessionStatus status;
+  InfcRequest* request;
+  xmlNodePtr xml;
+
+  g_return_val_if_fail(INFC_IS_SESSION_PROXY(proxy), NULL);
+  g_return_val_if_fail(n_params == 0 || params != NULL, NULL);
+
+  priv = INFC_SESSION_PROXY_PRIVATE(proxy);
+  g_return_val_if_fail(priv->session != NULL, NULL);
+
+  session_class = INF_SESSION_GET_CLASS(priv->session);
+
+  /* Make sure we are subscribed */
+  g_object_get(G_OBJECT(priv->session), "status", &status, NULL);
+  g_return_val_if_fail(status == INF_SESSION_RUNNING, NULL);
+  g_return_val_if_fail(priv->connection != NULL, NULL);
+  g_return_val_if_fail(priv->request_manager != NULL, NULL);
+
+  request = infc_request_manager_add_request(
+    priv->request_manager,
+    INFC_TYPE_REQUEST,
+    "user-join",
+    G_CALLBACK(func),
+    user_data,
+    NULL
+  );
+
+  xml = infc_session_proxy_request_to_xml(INFC_REQUEST(request));
+
+  g_assert(session_class->set_xml_user_props != NULL);
+  session_class->set_xml_user_props(priv->session, params, n_params, xml);
+
+  inf_communication_group_send_message(
+    INF_COMMUNICATION_GROUP(priv->subscription_group),
+    priv->connection,
+    xml
+  );
+
+  return INF_REQUEST(request);
 }
 
 /*
@@ -975,18 +1080,6 @@ infc_session_proxy_class_init(gpointer g_class,
 
   proxy_class->translate_error = infc_session_proxy_translate_error_impl;
 
-  g_object_class_install_property(
-    object_class,
-    PROP_SESSION,
-    g_param_spec_object(
-      "session",
-      "Session",
-      "The underlaying session object",
-      INF_TYPE_SESSION,
-      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY
-    )
-  );
-
   g_object_class_install_property(
     object_class,
     PROP_SUBSCRIPTION_GROUP,
@@ -1010,6 +1103,8 @@ infc_session_proxy_class_init(gpointer g_class,
       G_PARAM_READABLE
     )
   );
+
+  g_object_class_override_property(object_class, PROP_SESSION, "session");
 }
 
 static void
@@ -1024,6 +1119,16 @@ infc_session_proxy_communication_object_init(gpointer g_iface,
   iface->received = infc_session_proxy_communication_object_received;
 }
 
+static void
+infc_session_proxy_session_proxy_init(gpointer g_iface,
+                                      gpointer iface_data)
+{
+  InfSessionProxyIface* iface;
+  iface = (InfSessionProxyIface*)g_iface;
+
+  iface->join_user = infc_session_proxy_session_proxy_join_user;
+}
+
 GType
 infc_session_proxy_get_type(void)
 {
@@ -1050,6 +1155,12 @@ infc_session_proxy_get_type(void)
       NULL
     };
 
+    static const GInterfaceInfo session_proxy_info = {
+      infc_session_proxy_session_proxy_init,
+      NULL,
+      NULL
+    };
+
     session_proxy_type = g_type_register_static(
       G_TYPE_OBJECT,
       "InfcSessionProxy",
@@ -1062,6 +1173,12 @@ infc_session_proxy_get_type(void)
       INF_COMMUNICATION_TYPE_OBJECT,
       &communication_object_info
     );
+
+    g_type_add_interface_static(
+      session_proxy_type,
+      INF_TYPE_SESSION_PROXY,
+      &session_proxy_info
+    );
   }
 
   return session_proxy_type;
@@ -1176,86 +1293,6 @@ infc_session_proxy_set_connection(InfcSessionProxy* proxy,
   g_object_thaw_notify(G_OBJECT(proxy));
 }
 
-/**
- * infc_session_proxy_join_user:
- * @proxy: A #InfcSessionProxy.
- * @params: Construction properties for the InfUser (or derived) object.
- * @n_params: Number of parameters.
- * @error: Location to store error information.
- *
- * Requests a user join for a user with the given properties (which must not
- * include #InfUser:id and #InfUser:flags since these are initially set by the
- * server). The #InfUser:status property is optional and defaults to
- * %INF_USER_ACTIVE if not given. It must not be %INF_USER_UNAVAILABLE.
- *
- * Return Value: A #InfcUserRequest object that may be used to get notified
- * when the request succeeds or fails.
- **/
-InfcUserRequest*
-infc_session_proxy_join_user(InfcSessionProxy* proxy,
-                             const GParameter* params,
-                             guint n_params,
-                             GError** error)
-{
-  InfcSessionProxyPrivate* priv;
-  InfSessionClass* session_class;
-  InfSessionStatus status;
-  InfcRequest* request;
-  xmlNodePtr xml;
-
-  g_return_val_if_fail(INFC_IS_SESSION_PROXY(proxy), NULL);
-  g_return_val_if_fail(params != NULL || n_params == 0, NULL);
-
-  priv = INFC_SESSION_PROXY_PRIVATE(proxy);
-  g_return_val_if_fail(priv->session != NULL, NULL);
-
-  session_class = INF_SESSION_GET_CLASS(priv->session);
-
-  /* Make sure we are subscribed */
-  g_object_get(G_OBJECT(priv->session), "status", &status, NULL);
-  g_return_val_if_fail(status == INF_SESSION_RUNNING, NULL);
-  g_return_val_if_fail(priv->connection != NULL, NULL);
-  g_return_val_if_fail(priv->request_manager != NULL, NULL);
-
-  /* TODO: Check params locally */
-
-  request = infc_request_manager_add_request(
-    priv->request_manager,
-    INFC_TYPE_USER_REQUEST,
-    "user-join",
-    NULL
-  );
-
-  xml = infc_session_proxy_request_to_xml(INFC_REQUEST(request));
-
-  g_assert(session_class->set_xml_user_props != NULL);
-  session_class->set_xml_user_props(priv->session, params, n_params, xml);
-
-  inf_communication_group_send_message(
-    INF_COMMUNICATION_GROUP(priv->subscription_group),
-    priv->connection,
-    xml
-  );
-
-  return INFC_USER_REQUEST(request);
-}
-
-/**
- * infc_session_proxy_get_session:
- * @proxy: A #InfcSessionProxy.
- *
- * Returns the session proxied by @proxy, or %NULL if the session has been
- * closed.
- *
- * Return Value: A #InfSession, or %NULL.
- **/
-InfSession*
-infc_session_proxy_get_session(InfcSessionProxy* proxy)
-{
-  g_return_val_if_fail(INFC_IS_SESSION_PROXY(proxy), NULL);
-  return INFC_SESSION_PROXY_PRIVATE(proxy)->session;
-}
-
 /**
  * infc_session_proxy_get_connection:
  * @proxy: A #InfcSessionProxy.
@@ -1273,7 +1310,7 @@ infc_session_proxy_get_connection(InfcSessionProxy* proxy)
 }
 
 /**
- * infc_session_proxy_get_group:
+ * infc_session_proxy_get_subscription_group:
  * @proxy: A #InfcSessionProxy.
  *
  * Returns the group with all subscribed connections to the session, or %NULL
