@@ -1,5 +1,5 @@
 /* libinfinity - a GObject-based infinote implementation
- * Copyright (C) 2007-2011 Armin Burgmeier <armin@arbur.net>
+ * Copyright (C) 2007-2014 Armin Burgmeier <armin@arbur.net>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -34,8 +34,6 @@
 #include <libinfinity/communication/inf-communication-registry.h>
 #include <libinfinity/inf-signals.h>
 
-#include <libxml/xmlsave.h>
-
 typedef struct _InfCommunicationCentralMethodPrivate
   InfCommunicationCentralMethodPrivate;
 struct _InfCommunicationCentralMethodPrivate {
@@ -294,57 +292,31 @@ inf_communication_central_method_received(InfCommunicationMethod* method,
 {
   InfCommunicationCentralMethodPrivate* priv;
   InfCommunicationObject* target;
-  GError* error;
+  InfCommunicationGroup* group;
   InfCommunicationScope scope;
-  xmlBufferPtr buffer;
-  xmlSaveCtxtPtr ctx;
   gchar* remote_id;
   gchar* publisher_id;
   GSList* item;
 
   priv = INF_COMMUNICATION_CENTRAL_METHOD_PRIVATE(method);
-  error = NULL;
-
   target = inf_communication_group_get_target(priv->group);
+  group = priv->group;
+
   if(target != NULL)
   {
+    /* Make sure we stay alive, so we can check our connection list and
+     * publisher flag after the callback. This means we also need to keep
+     * group and target alive, to ensure the correct deinitialization order */
+    g_object_ref(target);
+    g_object_ref(group);
+    g_object_ref(method);
+
     scope = inf_communication_object_received(
       target,
       connection,
-      xml,
-      &error
+      xml
     );
 
-    /* TODO: Find a better place to show this message, maybe in
-     * InfCommunicationRegistry's received_cb. */
-    if(error != NULL)
-    {
-      buffer = xmlBufferCreate();
-      ctx = xmlSaveToBuffer(buffer, "UTF-8", XML_SAVE_FORMAT);
-      xmlSaveTree(ctx, xml);
-      xmlSaveClose(ctx);
-
-      g_object_get(G_OBJECT(connection), "remote-id", &remote_id, NULL);
-      publisher_id =
-        inf_communication_group_get_publisher_id(priv->group, connection);
-
-      g_warning(
-        "Error in group \"%s\", publisher \"%s\" while processing a message "
-        "from \"%s\":\n%s\n\nThe session might have lost consistency. The "
-        "message was:\n\n%s\n\n",
-        inf_communication_group_get_name(priv->group),
-        publisher_id,
-        remote_id,
-        error->message,
-        (const gchar*)xmlBufferContent(buffer)
-      );
-
-      g_free(publisher_id);
-      g_free(remote_id);
-      g_error_free(error);
-      xmlBufferFree(buffer);
-    }
-
     if(priv->is_publisher && scope == INF_COMMUNICATION_SCOPE_GROUP)
     {
       for(item = priv->connections; item != NULL; item = item->next)
@@ -361,6 +333,10 @@ inf_communication_central_method_received(InfCommunicationMethod* method,
       }
     }
 
+    g_object_unref(method);
+    g_object_unref(group);
+    g_object_unref(target);
+
     return scope;
   }
   else
