@@ -1,5 +1,5 @@
 /* libinfinity - a GObject-based infinote implementation
- * Copyright (C) 2007-2011 Armin Burgmeier <armin@arbur.net>
+ * Copyright (C) 2007-2014 Armin Burgmeier <armin@arbur.net>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -37,6 +37,7 @@
  **/
 
 #include <libinfinity/communication/inf-communication-registry.h>
+#include <libinfinity/communication/inf-communication-group-private.h>
 #include <libinfinity/common/inf-xml-util.h>
 #include <libinfinity/inf-signals.h>
 
@@ -555,14 +556,14 @@ inf_communication_registry_add_connection(InfCommunicationRegistry* registry,
 
     g_object_ref(connection);
 
-    g_signal_connect(
+    g_signal_connect_after(
       G_OBJECT(connection),
       "received",
       G_CALLBACK(inf_communication_registry_received_cb),
       registry
     );
 
-    g_signal_connect(
+    g_signal_connect_after(
       G_OBJECT(connection),
       "sent",
       G_CALLBACK(inf_communication_registry_sent_cb),
